from datetime import date
from typing import Any

from pandas import DataFrame


class Payslip:
    def __init__(
            self,
            payslip_name: str,
            payslip_date: date,
            worker_name: str,
            worker_id: str,
            additional_metadata: Any,
            payslip_records: DataFrame
    ) -> None:
        self.payslip_name = payslip_name
        self.payslip_date = payslip_date
        self.worker_name = worker_name
        self.worker_id = worker_id
        self.additional_metadata = additional_metadata
        self.monthly_records = payslip_records



import yaml

from payslip_parser.configuration.config import PayslipsParserConfig


def load_config(config_path: str) -> PayslipsParserConfig:
    with open(config_path) as f:
        return PayslipsParserConfig.parse_obj(
            yaml.safe_load(f)
        )



import abc
import os
import pathlib
from typing import List

import pandas as pd
from pandas import DataFrame

from payslip_parser.payslips.payslip import Payslip
from payslip_parser.parsers.base_payslip_parser import BasePayslipParser


class PayslipDirectoryParser(abc.ABC):
    """An ABC for parsing a directory of payslips of the same format"""

    def __init__(self, payslip_parser: BasePayslipParser):
        self.payslip_parser = payslip_parser

    def parse_directory(self, path: str) -> List[Payslip]:
        path = pathlib.Path(path)
        assert path.exists()
        assert path.is_dir()
        payslips_paths = [path / filename for filename in os.listdir(path) if filename.endswith('.pdf')]
        parsed_payslips = [self.payslip_parser.parse_payslip(str(p)) for p in payslips_paths]
        return parsed_payslips

    def directory_to_dataframe(self, path: str) -> DataFrame:
        parsed_payslips = self.parse_directory(path)
        payslip_dfs = []
        for payslip in parsed_payslips:
            records_df = payslip.monthly_records
            records_df = records_df.set_index(payslip.payslip_name + '__' + records_df.index.astype(str))
            records_df['payslip_name'] = payslip.payslip_name
            records_df['payslip_date'] = payslip.payslip_date
            records_df['worker_name'] = payslip.worker_name
            records_df['worker_id'] = payslip.worker_id
            payslip_dfs.append(records_df)
        return pd.concat(payslip_dfs)



from typing import Dict

from payslip_parser.text_blocks import BaseTextBlock


class IDFTextBlock(BaseTextBlock):
    # TODO consider implementing more elegantly (stop using dict as returned value)
    def _parse_text(
            self,
            raw_text,
            reverse_alphas=True
    ) -> Dict[str, str]:
        text = raw_text.strip('\n').replace('\xa0', ' ')

        alphas = ''
        numbers = ''
        for idx, char in enumerate(text):
            if char.isnumeric():
                numbers += char
            elif char == '.' and text[idx - 1].isnumeric() and text[idx + 1].isnumeric():
                numbers += char
            else:
                alphas = char + alphas if reverse_alphas else alphas + char

        return {
            'alphas': alphas,
            'numbers': numbers
        }



import re
from datetime import date
from typing import List, Dict, Any

from payslip_parser.parsers.base_payslip_parser import BasePayslipParser
from payslip_parser.text_blocks.base_text_block import BaseTextBlock
from payslip_parser.configuration.config import RegionBounds
from payslip_parser.text_blocks.idf_text_block import IDFTextBlock


class IDFPayslipParser(BasePayslipParser):
    def _get_payslip_name(self, payslip_path: str) -> str:
        filename = payslip_path.split('/')[-1].split('\\')[-1].split('.')[0]
        month = filename[8:-5] if len(filename[8:-5]) == 2 else '0' + filename[8:-5]  # add leading zero to month
        payslip_name = filename[:7] + '__' + filename[-5:-1] + '-' + month
        return payslip_name

    def _get_worker_id(self, header_blocks: List[BaseTextBlock]) -> str:
        idf_personal_number_block = header_blocks[4]
        assert idf_personal_number_block.region_name == 'personal_data'
        return idf_personal_number_block.parsed_text['numbers']

    def _get_worker_name(self, header_blocks: List[BaseTextBlock]) -> str:
        worker_name_block = header_blocks[0]
        assert worker_name_block.region_name == 'to'
        return worker_name_block.parsed_text['alphas'][4:]  # get rid of מ.א. prefix

    def _get_payslip_date(self, header_blocks: List[BaseTextBlock]) -> date:
        payslip_date_block = header_blocks[3]
        assert payslip_date_block.region_name == 'payslip_date'
        return date(
            year=int(payslip_date_block.parsed_text['numbers'][-4:]),
            month=int(payslip_date_block.parsed_text['numbers'][:-4]),
            day=1
        )

    def _get_additional_metadata(self, header_blocks: List[BaseTextBlock]) -> Any:
        payment_unit_block = header_blocks[2]
        assert payment_unit_block.region_name == 'to'

        national_id_block = header_blocks[5]
        assert national_id_block.region_name == 'personal_data'

        bank_name_block = header_blocks[6]
        account_details_block = header_blocks[7]
        assert bank_name_block.region_name == account_details_block.region_name == 'bank_details'

        return {
            'payment_unit': payment_unit_block.parsed_text['numbers'],  # יחתש
            'national_id': national_id_block.parsed_text['numbers'][:9],  # remove יחתש suffix, if exists
            'bank_details': {
                'bank_name': bank_name_block.parsed_text['alphas'],
                'branch_code': bank_name_block.parsed_text['numbers'],
                'branch_name': account_details_block.parsed_text['alphas'],
                'account_number': account_details_block.parsed_text['numbers']
            }
        }

    def _instantiate_text_block(self, block_id: int, region_bounds: RegionBounds, region_name: str, is_header: bool,
                                text: str) -> BaseTextBlock:
        return IDFTextBlock(block_id, region_bounds, region_name, is_header, text)

    def _body_block_to_record(self, block: IDFTextBlock) -> Dict:
        assert not block.is_header

        region_name = block.region_name
        alphas = block.parsed_text['alphas']
        numbers = block.parsed_text['numbers']

        if region_name == 'net_monthly_payment':
            return {'שכר חודשי נטו': float(numbers)}

        if region_name == 'net_differences':
            return {'הפרשים נטו': float(numbers)}

        if region_name == 'total_transfer_to_bank':
            return {'סהכ העברה לבנק': float(numbers)}

        if alphas == '':
            return {}

        if region_name == 'raw_data':
            return self._parse_raw_data_region(alphas, numbers)

        if region_name == 'monthly_payments':
            return self._parse_monthly_regions(alphas, numbers, negate_sums=False)

        if region_name == 'monthly_reductions':
            return self._parse_monthly_regions(alphas, numbers, negate_sums=True)

        if region_name == 'payments_differences':
            return self._parse_differences_regions(alphas, numbers, negate_total=False)

        if region_name == 'reductions_differences':
            return self._parse_differences_regions(alphas, numbers, negate_total=True)

    def _parse_raw_data_region(self, alphas: str, numbers: str) -> Dict:
        """Parses נתונים גולמיים"""
        # TODO improve this function
        if alphas.startswith('רגיל') or alphas.startswith('חובה'):
            start_date = date(
                day=int(numbers[:2]),
                month=int(numbers[2:4]),
                year=int(f'{self.config.century_prefix}{numbers[4:]}')
            )
            return {
                'שם הנתון': alphas[4:],
                'ערך הנתון': alphas[:4],
                'ת. תחילה': start_date,
            }

        if alphas.startswith('דחית שרות החובה'):
            start_date = date(
                day=int(numbers[:2]),
                month=int(numbers[2:4]),
                year=int(f'{self.config.century_prefix}{numbers[4:]}')
            )
            return {
                'שם הנתון': alphas[14:],
                'ערך הנתון': alphas[:14],
                'ת. תחילה': start_date,
            }

        return {
            'unknown': f'{alphas}{numbers}',
        }

    @staticmethod
    def _parse_monthly_regions(alphas: str, numbers: str, negate_sums: bool) -> Dict:
        """Parses תשלומים שוטפים or ניכויים שוטפים"""
        split_numbers = numbers.split('.')
        if len(split_numbers) == 2:  # only סכום נוכחי
            previous_sum = None
            current_sum = float(numbers)
        elif len(split_numbers) == 3:  # both סכום נוכחי and סכום קודם
            current_sum = float(split_numbers[0] + '.' + split_numbers[1][:2])  # shekels.agorot
            previous_sum = float(split_numbers[1][2:] + '.' + split_numbers[2])  # shekels.agorot
        else:
            raise RuntimeError(f'Error parsing {alphas}!')

        if negate_sums:  # negate sums if we are parsing ניכויים
            current_sum = -current_sum
            previous_sum = -previous_sum if previous_sum is not None else None

        return {
            'שם התשלום/ניכוי': alphas,
            'סכום קודם': previous_sum,
            'סכום נוכחי': current_sum,
        }

    def _parse_differences_regions(self, alphas: str, numbers: str, negate_total: bool) -> Dict:
        """Parses הפרשי תשלומים or הפרשי ניכויים"""
        pattern = '(?P<from_DD>[0-9]{2})(?P<from_MM>[0-9]{2})(?P<from_YY>[0-9]{2})' \
                  '(?P<to_DD>[0-9]{2})(?P<to_MM>[0-9]{2})(?P<to_YY>[0-9]{2})' \
                  '(?P<shekels>[0-9]{1,5}).(?P<agorot>[0-9]{2})'
        m = re.match(pattern, numbers)
        if m:
            from_date = date(
                year=int(f'{self.config.century_prefix}{m.group("from_YY")}'),
                month=int(m.group("from_MM")),
                day=int(m.group("from_DD"))
            )
            to_date = date(
                year=int(f'{self.config.century_prefix}{m.group("to_YY")}'),
                month=int(m.group("to_MM")),
                day=int(m.group("to_DD"))
            )
            total = float(f'{m.group("shekels")}.{m.group("agorot")}')
        else:
            raise RuntimeError(f'Error parsing {alphas}!')

        if negate_total:
            total = -total

        return {
            'שם הפרש תשלום/ניכוי': alphas,
            'תשלום/ניכוי מתאריך': from_date,
            'תשלום/ניכוי עד תאריך': to_date,
            'סכום הפרשי תשלומים/ניכויים': total,
        }



import abc
from datetime import date
from typing import Any, List, Dict

import fitz
from pandas import DataFrame

from payslip_parser import Payslip
from payslip_parser.text_blocks.base_text_block import BaseTextBlock
from payslip_parser.configuration.config import PayslipsParserConfig, RegionBounds


class BasePayslipParser(abc.ABC):
    """An ABC for parsing a single payslip"""

    def __init__(self, config: PayslipsParserConfig):
        self.config = config

    def parse_payslip(self, payslip_path: str) -> Payslip:
        # noinspection PyUnresolvedReferences
        pdf = fitz.open(payslip_path)
        text_blocks = []
        for page_idx, page in enumerate(pdf):
            if self.config.pages_to_ignore is not None and page_idx in self.config.pages_to_ignore:
                continue
            for pdf_block in page.get_textpage().extractBLOCKS():
                region_bounds = RegionBounds(
                    x0=pdf_block[0],
                    y0=pdf_block[1],
                    x1=pdf_block[2],
                    y1=pdf_block[3]
                )
                region_details = self._get_region_details(region_bounds)
                if region_details is not None:
                    region_name, is_header = region_details
                    text_block = self._instantiate_text_block(
                        block_id=pdf_block[5],
                        region_bounds=region_bounds,
                        region_name=region_name,
                        is_header=is_header,
                        text=pdf_block[4]
                    )
                    text_blocks.append(text_block)

        # header blocks are descriptive of the payslip in general, for example worker name
        header_blocks = [block for block in text_blocks if block.is_header]

        # body blocks will eventually become payslip records, for example base monthly salary
        body_blocks = [block for block in text_blocks if not block.is_header]

        return Payslip(
            payslip_name=self._get_payslip_name(payslip_path),
            payslip_date=self._get_payslip_date(header_blocks),
            worker_id=self._get_worker_id(header_blocks),
            worker_name=self._get_worker_name(header_blocks),
            additional_metadata=self._get_additional_metadata(header_blocks),
            payslip_records=self._get_payslip_records(body_blocks)
        )

    def _get_payslip_records(self, body_blocks: List[BaseTextBlock]) -> DataFrame:
        records = []
        for block in body_blocks:
            record = self._body_block_to_record(block)
            if record:
                records.append(record)
        df = DataFrame(records)
        df = df.dropna(how='all')
        df = df.reset_index(drop=True)
        return df

    def _get_region_details(self, bounds: RegionBounds) -> (str, bool):
        for region in self.config.regions:
            if bounds.x0 > region.bounds.x0 and \
                    bounds.y0 > region.bounds.y0 and \
                    bounds.x1 < region.bounds.x1 and \
                    bounds.y1 < region.bounds.y1:
                if region.is_ignore_region:
                    return None
                else:
                    return region.name, region.is_header
        raise RuntimeError(f'Unidentified region: {bounds}')

    @abc.abstractmethod
    def _get_payslip_name(self, payslip_path: str) -> str:
        pass

    @abc.abstractmethod
    def _get_worker_id(self, header_blocks: List[BaseTextBlock]) -> str:
        pass

    @abc.abstractmethod
    def _get_worker_name(self, header_blocks: List[BaseTextBlock]) -> str:
        pass

    @abc.abstractmethod
    def _get_payslip_date(self, header_blocks: List[BaseTextBlock]) -> date:
        pass

    def _get_additional_metadata(self, header_blocks: List[BaseTextBlock]) -> Any:
        pass

    @abc.abstractmethod
    def _instantiate_text_block(self, block_id: int, region_bounds: RegionBounds, region_name: str, is_header: bool,
                                text: str) -> BaseTextBlock:
        pass

    @abc.abstractmethod
    def _body_block_to_record(self, block: BaseTextBlock) -> Dict:
        pass



from typing import List, Optional

from pydantic import BaseModel


class RegionBounds(BaseModel):
    x0: float
    y0: float
    x1: float
    y1: float


class Region(BaseModel):
    name: str
    is_header: bool
    is_ignore_region: Optional[bool]
    bounds: RegionBounds


class PayslipsParserConfig(BaseModel):
    regions: List[Region]
    rtl_payslip: bool
    century_prefix: int
    pages_to_ignore: Optional[List[int]]



from setuptools import setup

setup(
    name='payslip_parser',
    version='0.1.1',
    description='Parse your payslips from PDFs to pandas dataframes.',
    author='Eldar Abraham',
    url='https://github.com/eldarab/payslip-parser',
    packages=[''],
    license='MIT',
    install_requires=[
        'setuptools',
        'pandas',
        'pathlib',
        'yaml',
        'pyyaml',
        'pydantic',
        'fitz>=1.22.3',
    ],
)



import abc
from typing import Union

from payslip_parser.configuration.config import RegionBounds


class BaseTextBlock(abc.ABC):
    def __init__(
            self,
            block_id: int,
            region_bounds: RegionBounds,
            region_name: Union[None, str],
            is_header: bool,
            text: str
    ) -> None:
        self.block_id = block_id
        self.region_bounds = region_bounds
        self.region_name = region_name
        self.is_header = is_header
        self.raw_text = text
        self.parsed_text = self._parse_text(self.raw_text)

    @abc.abstractmethod
    def _parse_text(self, *args, **kwargs):
        pass
